#include "DLSVMClassifier.h"

#define Malloc(type,n) (type *)malloc((n)*sizeof(type))

DLSVMClassifier::DLSVMClassifier()
{

}

DLSVMClassifier::DLSVMClassifier( string model_filename, bool bTesting )
{
	/* From libsvm documentation
	see the LIBSVM documentation at http://www.csie.ntu.edu.tw/~cjlin/libsvm/ for more details.
	options:
	-s svm_type : set type of SVM (default 0)
		0 -- C-SVC
		1 -- nu-SVC
		2 -- one-class SVM
		3 -- epsilon-SVR
		4 -- nu-SVR
	-t kernel_type : set type of kernel function (default 2)
		0 -- linear: u'*v
		1 -- polynomial: (gamma*u'*v + coef0)^degree
		2 -- radial basis function (RBF): exp(-gamma*|u-v|^2)
		3 -- sigmoid: tanh(gamma*u'*v + coef0)
	-d degree : set degree in kernel function (default 3)
	-g gamma : set gamma in kernel function (default 1/k)
	-r coef0 : set coef0 in kernel function (default 0)
	-c cost : set the parameter C of C-SVC, epsilon-SVR, and nu-SVR (default 1)
	-n nu : set the parameter nu of nu-SVC, one-class SVM, and nu-SVR (default 0.5)
	-p epsilon : set the epsilon in loss function of epsilon-SVR (default 0.1)
	-m cachesize : set cache memory size in MB (default 100)
	-e epsilon : set tolerance of termination criterion (default 0.001)
	-h shrinking: whether to use the shrinking heuristics, 0 or 1 (default 1)
	-b probability_estimates: whether to train a SVC or SVR model for probability estimates, 0 or 1 (default 0)
	-wi weight: set the parameter C of class i to weight*C, for C-SVC (default 1)

	The k in the -g option means the number of attributes in the input data.
	option -v randomly splits the data into n parts and calculates cross validation accuracy/mean squared error on them. */

	param.svm_type = C_SVC;
	param.kernel_type = RBF;
	param.degree = 2;
	param.gamma = 1;	// 1/k
	param.coef0 = 0;
	param.nu = 0.5;
	param.cache_size = 100;
	param.C = 100;
	param.eps = 1e-5;
	param.p = 0.1;
	param.shrinking = 1;
	param.probability = 0;
	param.nr_weight = 0;
	param.weight_label = NULL;
	param.weight = NULL;
    
	prob.l = 0;
	prob.x = 0;
	prob.y = 0;

	x_space = 0;
	
	model_file_name = model_filename;

	if( bTesting == true )
		model = svm_load_model(model_file_name.c_str());
}

DLSVMClassifier::~DLSVMClassifier()
{
	svm_destroy_param(&param);
	if(!prob.y)
		free(prob.y);
	if(!prob.y)
		free(prob.x);
	if(!x_space)
		free(x_space);
	svm_destroy_model(model);
}

int DLSVMClassifier::batchTraining(vector< vector<double> > data, vector<int> dataclass)
{
	// Check the input data first
	int NrOfSamples = int(data.size());
	int NrOfGroundtruth = int(dataclass.size());
	int dataLength = 0;

	if ( NrOfSamples >= 1 )
	{
		vector< vector<double> >::const_iterator data_Iter;
		dataLength = int(data.begin()->size());
		for (data_Iter = data.begin(); data_Iter != data.end(); data_Iter++)
		{
			if (data_Iter->size() != dataLength)
			{
				// mismatch in the lengths of data vectors
				throw DLException(DL_MEMORY_EXCEPTION, "<DLSVMClassifier::batchTraining()> Mismatch in the lengths of data vector!", __FILE__, __LINE__);
			}
		}
		if (NrOfSamples != NrOfGroundtruth)
		{
			// mismatch between the number of data samples and their ground truth labels
			throw DLException(DL_MEMORY_EXCEPTION, "<DLSVMClassifier::batchTraining()> Mismatch between the number of data samples and their ground truth labels!", __FILE__, __LINE__);
		}
	}
	else
	{
		throw DLException(DL_MEMORY_EXCEPTION, "<DLSVMClassifier::batchTraining()> Empty vector of class label!", __FILE__, __LINE__);
	}

	prob.l = NrOfSamples;
	int elements = NrOfSamples * (dataLength + 1 );

	prob.y = Malloc(double, prob.l);
	prob.x = Malloc(struct svm_node *, prob.l);
	x_space = Malloc(struct svm_node, elements);

	int idx = 0;
	int idx1 = 0;
	for(int i=0; i < NrOfSamples; i++)
	{
		prob.y[idx1] = dataclass[i];
		prob.x[idx1] = &x_space[idx];

		for( int j = 0; j < data[0].size(); j++)
		{
			x_space[idx].index = j;
			x_space[idx].value = data[i][j];
			idx++;
		}

		x_space[idx++].index = -1;

		idx1++;
	}

	if(param.gamma == 0)
	{
		sort(dataclass.begin(), dataclass.end());
		int maxVal = dataclass[dataclass.size()-1];
		param.gamma = 1.0 / maxVal;
	};

	error_msg = svm_check_parameter(&prob, &param);

	if(error_msg)
	{
		cout << error_msg << endl;
		return -1;
	}

	model = svm_train(&prob, &param);
	svm_save_model(model_file_name.c_str(), model);

	return 1;
}

vector< vector<double> > DLSVMClassifier::batchTesting(vector< vector<double> > data, vector<int> dataclass)
{
	// Check the input data first
	int NrOfSamples = int(data.size());
	int NrOfGroundtruth = int(dataclass.size());
	int dataLength = 0;

	if ( NrOfSamples >= 1 )
	{
		vector< vector<double> >::const_iterator data_Iter;
		dataLength = int(data.begin()->size());
		for (data_Iter = data.begin(); data_Iter != data.end(); data_Iter++)
		{
			if (data_Iter->size() != dataLength)
			{
				// mismatch in the lengths of data vectors
				throw DLException(DL_MEMORY_EXCEPTION, "<DLSVMClassifier::batchTraining()> Mismatch in the lengths of data vector!", __FILE__, __LINE__);
			}
		}
		if (NrOfSamples != NrOfGroundtruth)
		{
			// mismatch between the number of data samples and their ground truth labels
			throw DLException(DL_MEMORY_EXCEPTION, "<DLSVMClassifier::batchTraining()> Mismatch between the number of data samples and their ground truth labels!", __FILE__, __LINE__);
		}
	}
	else
	{
		throw DLException(DL_MEMORY_EXCEPTION, "<DLSVMClassifier::batchTraining()> Empty vector of class label!", __FILE__, __LINE__);
	}
 
	int nr_class = svm_get_nr_class(model);

	vector<int> testPredict;
	struct svm_node *x;
	x = (struct svm_node *) malloc( (data[0].size()  + 1)*sizeof(struct svm_node));

	int idx1 = 0;
	for(int i=0;i< data.size(); i++)
	{
		for( int j = 0; j < data[0].size(); j++)
		{
			x[j].index = j;
			x[j].value = data[i][j];
		}

		x[data[0].size()].index = -1;
		int v = svm_predict(model, x);
		testPredict.push_back(v);

		idx1++;
	}

	// creating confusion matrix
	for( int i = 0; i < nr_class; i++)
	{
		vector< double > tmp;
		for( int i = 0; i < nr_class; i++)
		{
			tmp.push_back(0);
		}
		confusionMatrix.push_back(tmp);
	}

	for( int i = 0; i < dataclass.size(); i++) {
		confusionMatrix[dataclass[i] - 1][testPredict[i] - 1] ++;
	}

	for( int i = 0; i < confusionMatrix.size(); i++)
	{	
		int sumval = 0;
		for( int j = 0; j < confusionMatrix.size(); j++) {
			sumval += confusionMatrix[i][j];
		}

		for( int j = 0; j < confusionMatrix.size(); j++) {
			confusionMatrix[i][j] = confusionMatrix[i][j]/sumval;
		}
	}

	return confusionMatrix;
}


int DLSVMClassifier::test(vector<double> testdata)
{
	struct svm_node *x;
	x = (struct svm_node *) malloc( (testdata.size()  + 1)*sizeof(struct svm_node));
	for( int j = 0; j < int(testdata.size()); j++)
	{
		x[j].index = j;
		x[j].value = testdata[j];
	}

	x[testdata.size()].index = -1;
	int iclass = int(svm_predict(model, x));

	return iclass;
}

int DLSVMClassifier::saveConfusionMatrix(string strFileName)
{
	FILE* f = fopen( strFileName.c_str(), "w+" );

	char pchElement[100];
	for(unsigned int i = 0; i < confusionMatrix.size(); i++) 
	{
		for(unsigned int j = 0; j < confusionMatrix.size()-1; j++) 
		{
			memset( pchElement, 0, 100 );
			sprintf(pchElement, "%4.4lf\t", confusionMatrix[i][j]);
			fwrite( pchElement, 1, strlen(pchElement), f  );
		}

		memset( pchElement, 0, 100 );
		sprintf(pchElement, "%4.4lf\n", confusionMatrix[i][confusionMatrix.size()-1]);
		fwrite( pchElement, 1, strlen(pchElement), f  );
	}
	fclose( f );

	return 1;
}
