#ifndef _DLSVMCLASSIFIER_H_
#define _DLSVMCLASSIFIER_H_

#include "svm.h"
#include <string>
#include <vector>
#include <iostream>
#include <algorithm>
#include "DLException.h"

using namespace std;

/**
 * \brief The DLSVMClassifier object is a wrapper class that provides an easy-to-use interface
 * for the original C implementation of LIBSVM. In addition, it include member functions for
 * batch training and test.
 *
 * LIBSVM is an integrated library for multi-class support vector classification (C-SVC, nu-SVC), 
 * regression (epsilon-SVR, nu-SVR) and distribution estimation (one-class SVM). 
 * The LIBSVM code here is based on LIBSVM release v2.85. Please see the LIBSVM website at 
 * http://www.csie.ntu.edu.tw/~cjlin/libsvm/ for documentations and updates.
 * 
 * Features currently supported in this verison of LIBSVM library include:
 *  - Different SVM formulations 
 *  - Efficient multi-class classification 
 *  - Cross validation for model selection 
 *  - Probability estimates 
 *  - Weighted SVM for unbalanced data 
 *  - Both C++ and Java sources
 *  - Support Python, R (also Splus), MATLAB, Perl, Ruby, Weka, CLISP and LabVIEW interfaces
*/

class DLSVMClassifier
{
public:

	/**
	* Default DLSVMClassifier Constructor.
	*/
	DLSVMClassifier();

	/**
	* DLSVMClassifier Constructor: Initialize SVM parameters and perform either training or test based on the input parameters
	* In training, a new model file in LIBSVM format will be created by the given file name. 
	* In testing, the specified model file in LIBSVM format will be used for classification.
	* \param model_filename file name of the SVM model file in LIBSVM format.
	* \param bTesting boolean indicator. True for testing; False for training.
	*/
	DLSVMClassifier(string model_filename, bool bTesting = true);

	/**
	* Default DLSVMClassifier Destructor: Free memory for LIBSVM.
	*/
	~DLSVMClassifier();

	/**
	* Batch training using the data and ground truth supplied
	* \param data input training data in the form of vector of vector. The dimension of data must
	* match the lengths of ground truth class labels.
	* \param dataclass ground truth information of the true class label of the data. 
	* Each class label must be distinct integer values.
	* \return 1 if successful 
	*/
	int batchTraining(vector< vector<double> > data, vector<int> dataclass);

	/**
	* Batch testing for evaluation using the data and ground truth supplied
	* \param data input test data in the form of vector of vector. The dimension of data must
	* match the lengths of ground truth class labels.
	* \param dataclass ground truth information of the true class label of the data. 
	* Each class label must be distinct integer values.
	* \return entries in the confusion matrix in the form of vector of vector.
	*/
	vector< vector<double> > batchTesting(vector< vector<double> > data, vector<int> dataclass);

	/**
	* Perform classification on a novel pattern represented by the vector of test data using the current model
	* \param testdata the test data correponding to the novel pattern to be classified using the current model
	* \return the predicted class label of the test data using the current model
	*/
	int test(vector<double> testdata);

	/**
	* Save the confusion matrix in the specified file
	* \param strFileName name of the file to be used to store the confusion matrix
	* \return 1 if successful
	*/
	int saveConfusionMatrix(string strFileName);

private:

	/// SVM parameter set
	struct svm_parameter param;

	/// The followings are the internal structs used by LIBSVM
	struct svm_problem prob;
	struct svm_model *model;
	struct svm_node *x_space;
	int cross_validation;
	int nr_fold;
	const char *error_msg;

	string model_file_name;

	/// Entries of the confusion matrix
	vector< vector<double> > confusionMatrix;
};

#endif